from swarmtorch.hybrid.model_training.eho import EHO, ChickenSwarm, SMA
from swarmtorch.hybrid.model_training.cat import (
    CatSwarm,
    Cockroach,
    Coati,
    Gorilla,
    GorillaTroopsOptimizer,
    JSO,
    KHA,
)

__all__ = [
    "EHO",
    "ChickenSwarm",
    "SMA",
    "CatSwarm",
    "Cockroach",
    "Coati",
    "Gorilla",
    "JSO",
    "KHA",
]
